/*
 * linux/drivers/char/psaux.c
 *
 * Driver for PS/2 type mouse by Johan Myreen.
 *
 * Supports pointing devices attached to a PS/2 type
 * Keyboard and Auxiliary Device Controller.
 *
 * Corrections in device setup for some laptop mice & trackballs.
 * 02Feb93  (troyer@saifr00.cfsat.Honeywell.COM,mch@wimsey.bc.ca)
 *
 * Changed to prevent keyboard lockups on AST Power Exec.
 * 28Jul93  Brad Bosch - brad@lachman.com
 *
 * Modified by Johan Myreen (jem@pandora.pp.fi) 04Aug93
 *   to include support for QuickPort mouse.
 *
 * Changed references to "QuickPort" with "82C710" since "QuickPort"
 * is not what this driver is all about -- QuickPort is just a
 * connector type, and this driver is for the mouse port on the Chips
 * & Technologies 82C710 interface chip. 15Nov93 jem@pandora.pp.fi
 *
 * Added support for SIGIO. 28Jul95 jem@pandora.pp.fi
 *
 * Rearranged SIGIO support to use code from tty_io.  9Sept95 ctm@ardi.com
 *
 * Modularised 8-Sep-95 Philip Blundell <pjb27@cam.ac.uk>
 *
 * Fixed keyboard lockups at open time
 * 3-Jul-96, 22-Aug-96 Roman Hodek <Roman.Hodek@informatik.uni-erlangen.de>
 */

/*
 * Kernel functions used by this driver:
 *
 * TODO: how do we check these?  get_user takes any pointer to a
 * non-void basic type and returns one value of that basic type.
 * put_user takes both the datum and the pointer, writing the datum
 * into the address specified by the pointer.
 * 
 *   get_user - macro 
 *   put_user - macro
 *
 * For save_flags, restore_flags, cli, and sti, we may want to say *
 * that any function that calls cli must either: (1) be able to prove
 * that it was previously set and also call sti before function exit,
 * or (2) be wrapped in a correctly coded save_flags, restore_flags
 * pair.  More succinctly, we could say that all (top-level?)
 * functions must leave the interrupt flag as they found it.  Also,
 * all computation that occurs in "interrupt time" must be bounded.
 *
 *   save_flags - macro
 *     - stores the current flags in the given unsigned long
 *
 *   restore_flags - macro
 *     - restores the current flags from the given unsigned long
 *
 *   cli - macro
 *     - clears the interrupt flag (disabling interrupts)
 *
 *   sti - macro
 *     - sets the interrupt flag (enabling interrupts)
 *
 * TODO: My book is not clear about how I/O addresses are specified by
 * a driver, but they must be on account of /proc/ioports.  Anyway, we
 * should check that the ioport specified by each of these be within
 * that range.  Also, look into probing more.
 *
 *   inb
 *   outb_p - pausing outb for slow devices
 *   inb_p - pausing inb for slow devices
 *
 * There are lots of things we could check about these two, but I
 * don't think any of them matter as far as a driver being able to
 * crash the kernel.  We should probably just check the usual stuff
 * (types and memory safety).
 *
 *   kill_fasync
 *   fasync_helper
 *
 * Nothing to check here.
 *
 *   add_mouse_randomness
 *
 * Ok, so these guys may be hard.  Essentially, we need to check that
 * we don't free any IRQs that we don't first successfully request.
 * Also, we may want to check that all requested IRQs are eventually
 * freed.  (This is much less important except in a system with PCMCIA
 * cards.)  We need to check that the dev_id passed in request_irq is
 * the same as the one passed in the corresponding free_irq, and the
 * values of the flags in request_irq determine whether the handler is
 * called with interrupts enabled or not.  If SA_INTERRUPT is not
 * passed in as a flag, the handler will be run with interrupts
 * enabled, and thus can take all the time that it wants (eventually
 * bounded) to finish.  Otherwise, it will run in interrupt time and
 * must behave accordingly.
 *
 *   free_irq
 *   request_irq
 *
 * wait must point into stack frame
 * must call add and remove in pairs
 * race condition? (p. 209)
 * 
 *   wake_up_interruptible
 *   add_wait_queue
 *   remove_wait_queue
 *
 * cannot be called during interrupt time
 *
 *   schedule
 *
 *   select_wait
 *   misc_register
 *   misc_deregister
 *   printk
 *
 * If flags are GFP_KERNEL or GFP_NFS, kmalloc may sleep, so callers
 * should be reentrant (or only called once, as in psaux_init, below).
 * If kmalloc is called at interrupt time, it should use GFP_ATOMIC to
 * avoid being put to sleep.
 *
 *   kmalloc
 *   kfree 
 */

/*
 * Kernel Entry Points:
 *
 * General Notes:
 *
 *   (1) Except where noted, all functions are called with interrupts
 *   enabled, and are expected to return in that same state.
 * 
 * Essentially, whatever open does, release should undo.  Generally,
 * there can be more than one process with this device open at a time,
 * so we will keep some count of the number of them (aux_count).  The
 * first open will call request_irq and MOD_INC_USE_COUNT (and maybe
 * kmalloc), and the last release will call free_irq and
 * MOD_DEC_USE_COUNT (and maybe kfree).  release should also be able
 * to assume that the device is open (TODO: is this correct?).
 *
 *   open
 *   release
 *
 * The interaction between these two is similar to that of open and
 * release.  Paired functions are misc_register and misc_deregister,
 * kmalloc and kfree.  Neither of these should involve request_irq or
 * free_irq, however, so that two modules can share an irq line as
 * long as they are not simultaneously open (they could still share
 * using SA_SHARED, if they wanted).
 *
 *   init_module
 *   cleanup_module
 *
 * There's really nothing special about read and write, other than
 * that they should obey the POSIX conventions as far as their pre and
 * post conditions go.  They should both be able to assume the device
 * has been openned (TODO: is this correct?).
 *
 *   read
 *   write
 *
 * This is the only entry point that I know of that can be called
 * during interrupt time (ie., with interrupts disabled), and even
 * then only if the interrupt it is servicing was registered with the
 * SA_INTERRUPT flag.  If the function is called in interrupt time, it
 * must return in interrupt time as well.  The only concrete thing to
 * check about this function is that it not call any other functions
 * which require that interrupts be enabled.
 *
 *   interrupt
 *
 * These two are not that exciting, and for our first pass I think we
 * can more or less ignore them.
 *
 *   fasync
 *   select 
 */

/*
 * HARDWARE NOTES
 *
 * From http://www.cubic.org/source/archive/hardware/ports/ports.txt:
 *
 * "A read from port 64H gives the following status byte:
 *
 *   Bit     Function
 *    7      1 = Parity error
 *    6      1 = General Time Out
 *    5      1 = Auxiliary output buffer full
 *    4      1 = Inhibit switch
 *    3      1 = Command/data
 *    2      1 = System flag
 *    1      1 = Input buffer full
 *    0      1 = Output buffer full
 *
 * The status register can be read at any time.  The data port 60H
 * should be read only when the output buffer full bit in the status
 * register is 1. Data should be written to the 8042 input buffer only
 * when the input buffer full bit in the status register is 0.  If the
 * auxiliary output buffer full bit is 1, then the data read came from
 * the auxiliary device.  The command port 64H should be written to
 * only when the status register input buffer full bit and the output
 * buffer full bit are 0.  Devices connected to the 8042 should be
 * disabled before sending a command that generates output."
 *
 * In the context of this file, we are the "auxiliary device".
 * Continuing:
 *
 * 
 */

/* 
 * Notes about the keyboard bottom half (KEYBOARD_BH):
 *
 * The keyboard bottom half sends data to address 0x60, which is
 * shared with us.  Because of this, any of our functions that don't
 * want what they are doing to be interferred with by the keyboard
 * driver will call (en/dis)able_bh (KEYBOARD_BH) around their code.  
 */

/* Uncomment the following line if your mouse needs initialization. */

/* #define INITIALIZE_DEVICE */

#include <linux/module.h>
 
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/fcntl.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <linux/malloc.h>
#include <linux/miscdevice.h>
#include <linux/random.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

#include <linux/config.h>

#define PSMOUSE_MINOR      1            /* minor device # for this mouse */

/* aux controller ports */
#define AUX_INPUT_PORT	0x60		/* Aux device output buffer */
#define AUX_OUTPUT_PORT	0x60		/* Aux device input buffer */
#define AUX_COMMAND	0x64		/* Aux device command buffer */
#define AUX_STATUS	0x64		/* Aux device status reg */

/* aux controller status bits */
#define AUX_OBUF_FULL	0x21		/* output buffer (from device) full */
#define AUX_IBUF_FULL	0x02		/* input buffer (to device) full */

/* aux controller commands */

/* If we write a value of 0x60 to port AUX_COMMAND, then the next
 * value written to port AUX_OUTPUT_PORT will be placed in the 8042
 * command register.  The flags AUX_INTS_ON and AUX_INTS_OFF (defined
 * below) are appropriate for this purpose. --srhea 
 */

#define AUX_CMD_WRITE	0x60		/* value to write to controller */

/* 
 * If we write a value of 0xd4 to port AUX_COMMAND, then the next
 * value written to port AUX_OUTPUT_PORT will go to the "auxiliary
 * device" (ie., the mouse)
 */

#define AUX_MAGIC_WRITE	0xd4		/* value to send aux device data */

#define AUX_INTS_ON	0x47		/* enable controller interrupts */
#define AUX_INTS_OFF	0x65		/* disable controller interrupts */

/* 
 * These two flags work by either driving the clock line low
 * (disable), or pulsing it as usual (enable). --srhea 
 */

#define AUX_DISABLE	0xa7		/* disable aux */
#define AUX_ENABLE	0xa8		/* enable aux */

/* aux device commands */
#define AUX_SET_RES	0xe8		/* set resolution */
#define AUX_SET_SCALE11	0xe6		/* set 1:1 scaling */
#define AUX_SET_SCALE21	0xe7		/* set 2:1 scaling */
#define AUX_GET_SCALE	0xe9		/* get scaling factor */
#define AUX_SET_STREAM	0xea		/* set stream mode */
#define AUX_SET_SAMPLE	0xf3		/* set sample rate */

#define AUX_ENABLE_DEV	0xf4		/* enable aux device */
#define AUX_DISABLE_DEV	0xf5		/* disable aux device */
#define AUX_RESET	0xff		/* reset aux device */

#define MAX_RETRIES	60		/* some aux operations take long time*/
#if defined(__alpha__) && !defined(CONFIG_PCI)
# define AUX_IRQ	9		/* Jensen is odd indeed */
#else
# define AUX_IRQ	12
#endif
#define AUX_BUF_SIZE	2048

extern unsigned char aux_device_present;
extern unsigned char kbd_read_mask;	/* from keyboard.c */

/* The structure defining the queue which we use to hold data coming
   from the mouse port. --srhea */

struct aux_queue {
	unsigned long head;
	unsigned long tail;

        /* Holds the list of any processes asleep on this device.  See
           linux/sched.h. --srhea */

	struct wait_queue *proc_list;
	struct fasync_struct *fasync;
	unsigned char buf[AUX_BUF_SIZE];
};

static struct aux_queue *queue;

/* Is there any data ready in the queue? --srhea */

static int aux_ready = 0;

/* How many people have us open at the moment? --srhea */
static int aux_count = 0;

static int aux_present = 0;
static int poll_aux_status(void);
static int poll_aux_status_nosleep(void);
static int fasync_aux(struct inode *inode, struct file *filp, int on);


/*
 * Write to aux device 
 *
 * Sends the value `val' to the aux device.  (This function is only
 * called from open_aux, below.) --srhea 
 */

static void aux_write_dev(int val)
{
	poll_aux_status();
	outb_p(AUX_MAGIC_WRITE,AUX_COMMAND);	/* write magic cookie */
	poll_aux_status();
	outb_p(val,AUX_OUTPUT_PORT);		/* write data */
}

/*
 * Write to device & handle returned ack
 *
 * This function is only called from psaux_init, below. --srhea 
 */

#if defined INITIALIZE_DEVICE
static int aux_write_ack(int val)
{
	int retries = 0;

	poll_aux_status_nosleep();
	outb_p(AUX_MAGIC_WRITE,AUX_COMMAND);
	poll_aux_status_nosleep();
	outb_p(val,AUX_OUTPUT_PORT);
	poll_aux_status_nosleep();

	if ((inb(AUX_STATUS) & AUX_OBUF_FULL) == AUX_OBUF_FULL)
	{
		return (inb(AUX_INPUT_PORT));
	}
	return 0;
}
#endif /* INITIALIZE_DEVICE */

/*
 * Write aux device command
 */

static void aux_write_cmd(int val)
{
	poll_aux_status();
	outb_p(AUX_CMD_WRITE,AUX_COMMAND);
	poll_aux_status();
	outb_p(val,AUX_OUTPUT_PORT);
}

/*
 * This is an atomic queue pop operation on the aux_queue. --srhea
 */

static unsigned int get_from_queue(void)
{
	unsigned int result;
	unsigned long flags;

	save_flags(flags);
	cli();
	result = queue->buf[queue->tail];
	queue->tail = (queue->tail + 1) & (AUX_BUF_SIZE-1);
	restore_flags(flags);
	return result;
}


static inline int queue_empty(void)
{
	return queue->head == queue->tail;
}



/*
 * Interrupt from the auxiliary device: a character
 * is waiting in the keyboard/aux controller.
 *
 * TODO: why doesn't this function call disable_bh (KEYBOARD_BH), like
 * open_aux (and the others)?  It's registered as a slow interrupt, so
 * it should run with interrupts enabled, right? --srhea
 */

static void aux_interrupt(int cpl, void *dev_id, struct pt_regs * regs)
{
	int head = queue->head;
	int maxhead = (queue->tail-1) & (AUX_BUF_SIZE-1);

	/* If the device's output buffer is not full, don't read from
           it yet.  TODO: Why? --srhea */
	
	if ((inb(AUX_STATUS) & AUX_OBUF_FULL) != AUX_OBUF_FULL)
		return;

	/* Store the new data into the head of the queue.  Also, give
           our data to the kernel to use in random number generation.
           See random.h. --srhea */

	add_mouse_randomness(queue->buf[head] = inb(AUX_INPUT_PORT));

	/* If the queue is not yet full, increment the head
           pointer. --srhea */

	if (head != maxhead) {
		head++;
		head &= AUX_BUF_SIZE-1;
	}
	queue->head = head;

	/* There is now data ready in the queue */

	aux_ready = 1;

	/* Signal any asynchronous readers (not blocked readers) that
           data has arrived.  An example of such a process is X (I
           think?).  --srhea */

	if (queue->fasync)
		kill_fasync(queue->fasync, SIGIO);

	/* Wake up any processes that are blocked on this
           device. --srhea */

	wake_up_interruptible(&queue->proc_list);
}

/* 
 * The "close" method for this device. --srhea
 *
 * Postcondition: (aux_count > 0) || (free_irq && MOD_DEC_USE_COUNT)
 */

static void release_aux(struct inode * inode, struct file * file)
{
	/* Remove this file from the list of asynchronous readers, if
	   it is one. -- srhea */

	fasync_aux(inode, file, 0);

	/* We now have one less user, so decrement the aux_count.  If
           it is still positive, however, the device is still in use,
           so we don't need to do anything more. --srhea */

	if (--aux_count)
		return;

	/* Otherwise, we need to clean up after ourselves. --srhea */

	/* This function runs with interrupts enabled, so it's
           possible that the keyboard bottom half could be run at any
           time.  Thus we must disable it to avoid mixing of command
           bytes. --srhea */

	disable_bh(KEYBOARD_BH);

	/* Since we are going away, there won't be anyone left to
           handle any aux interrupts, so we must disable them. --srhea */

	aux_write_cmd(AUX_INTS_OFF);		/* disable controller ints */
	poll_aux_status();

	outb_p(AUX_DISABLE,AUX_COMMAND);      	/* Disable Aux device */
	poll_aux_status();

	/* reenable kbd bh */
	enable_bh(KEYBOARD_BH);

	free_irq(AUX_IRQ, NULL);
	MOD_DEC_USE_COUNT;
}

/* 
 * Call fasync_helper on with the given arguments and a pointer to our
 * fasync queue.  Return 0 on success. --srhea
 */

static int fasync_aux(struct inode *inode, struct file *filp, int on)
{
	int retval;

	retval = fasync_helper(inode, filp, on, &queue->fasync);
	if (retval < 0)
		return retval;
	return 0;
}

/*
 * Install interrupt handler.
 * Enable auxiliary device.
 *
 * Postcondition: (aux_count <= 1) || (request_irq && MOD_INC_USE_COUNT)
 */

static int open_aux(struct inode * inode, struct file * file)
{
	/* TODO: what does this do? --srhea */

	if (!aux_present)
		return -ENODEV;

	/* Increase the count of the number of people who have openned
	   the device.  If it was already non-zero, then the device
	   has already been openned, and we are done. --srhea */

	if (aux_count++)
		return 0;

	/* Is the device ready to be sent commands?  If not,
           fail. --srhea */

	if (!poll_aux_status()) {
		aux_count--;
		return -EBUSY;
	}

	/* Flush input queue */

	queue->head = queue->tail = 0;

	/* Register our IRQ.  Note that aux_interrupt is registered as
           a slow interrupt handler (because the SA_INTERRUPT flag is
           not set).  TODO: Why? --srhea */

	if (request_irq(AUX_IRQ,          /* IRQ number */
			aux_interrupt,    /* handler function pointer */
			0,                /* flags */
			"PS/2 Mouse",     /* device name */
			NULL)) {          /* device id */
		aux_count--;
		return -EBUSY;
	}
	MOD_INC_USE_COUNT;

	/* This function runs with interrupts enabled, so it's
           possible that the keyboard bottom half could be run at any
           time.  Thus we must disable it to avoid mixing of command
           bytes. --srhea */

	disable_bh(KEYBOARD_BH);
	poll_aux_status();
	outb_p(AUX_ENABLE,AUX_COMMAND);		/* Enable Aux */
	aux_write_dev(AUX_ENABLE_DEV);		/* enable aux device */
	aux_write_cmd(AUX_INTS_ON);		/* enable controller ints */
	poll_aux_status();
	/* reenable kbd bh */
	enable_bh(KEYBOARD_BH);

	/* Since we've just been openned, there is obviously no data
           ready. --srhea */

	aux_ready = 0;
	return 0;
}

/*
 * Write to the aux device.
 *
 * inode  - pointer to the inode of the device in the file system
 * file   - file structure for the device
 * buffer - array of bytes to be written
 * count  - number of bytes to write
 *
 * Postcondition: result <= count
 *
 */

static int write_aux(struct inode * inode, struct file * file, 
		     const char * buffer, int count)
{
	int retval = 0;

	/* If count is zero, return zero. --srhea */

	if (count > 0) {
		int written = 0;

		/* disable kbd bh to avoid mixing of cmd bytes */
		disable_bh(KEYBOARD_BH);

		/* This loop has a built in bound, since count is
                   originally greater than zero, is decremented by one
                   every cycle, and causes the loop to terminate when
                   it reaches zero. --srhea */

		do {
                        /* If the device is not ready, fall
                           out. --srhea */

			if (!poll_aux_status())
				break;

			/* Tell the device a write is comming. --srhea */

			outb_p(AUX_MAGIC_WRITE,AUX_COMMAND);

			/* If the device is not ready, fall
                           out. --srhea */

			if (!poll_aux_status())
				break;

			/* Write a data byte.  get_user translates the
                           virtual address in buffer to the
                           corresponding physical address, and returns
                           one datum of the type of &buffer.  See
                           asm/segment.h. --srhea */

			outb_p(get_user(buffer++),AUX_OUTPUT_PORT);

			/* Keep track of the number of bytes
                           writen. --srhea */

			written++;
		} while (--count);

		/* reenable kbd bh */
		enable_bh(KEYBOARD_BH);

		/* If we didn't write any bytes at all, and count was
                   greater than zero, fail by returning EIO.
                   Otherwise, return the number of bytes written, and
                   increment the inode's last modification time to the
                   current time. --srhea */

		retval = -EIO;
		if (written) {
			retval = written;
			inode->i_mtime = CURRENT_TIME;
		}
	}

	return retval;
}


/*
 * Put bytes from input queue to buffer.
 *
 * inode  - pointer to the inode of the device in the file system
 * file   - file structure for the device
 * buffer - array of bytes to put data into
 * count  - maximum number of bytes to read
 *
 * Returns the actual number of bytes written into the buffer.
 *
 * Postcondition: result <= count
 *
 */

static int read_aux(struct inode * inode, struct file * file, 
		    char * buffer, int count)
{
	struct wait_queue wait = { current, /* current task */
				   NULL };  /* next item in the queue */
	int i = count;
	unsigned char c;

	/* If the queue is empty (we have no data to give this
           process), either put it to sleep or return EAGAIN,
           depending on whether we are in blocking mode or
           not. --srhea */

	if (queue_empty()) {
		if (file->f_flags & O_NONBLOCK)
			return -EAGAIN;
		
		/* Add this process to the wait queue. --srhea */

		add_wait_queue(&queue->proc_list, &wait);

		/* Set the state of the process so that we will be
                   awakened on interrupts. --srhea */
repeat:
		current->state = TASK_INTERRUPTIBLE;
		
		/* While there is still no data, and while no signals
                   have arrived for this process (such as a SIGINT,
                   perhaps), go to sleep. --srhea */

		if (queue_empty() && !(current->signal & ~current->blocked)) {
			schedule();
			goto repeat;
		}

		/* Wake the task up and take it off the wait
                   queue. --srhea */

		current->state = TASK_RUNNING;
		remove_wait_queue(&queue->proc_list, &wait);
	}

	/* Now if there is data to read, read as much as is available
           or i bytes worth, whichever is less. --srhea */

	while (i > 0 && !queue_empty()) {
		c = get_from_queue();
		put_user(c, buffer++);
		i--;
	}

	/* If the queue is not empty, there is still data
           ready. --srhea */

	aux_ready = !queue_empty();

	/* If we actually read any data into the buffer, set the last
           read time of the inode to the current time and return the
           number of bytes we read. */
      
	if (count-i) {
		inode->i_atime = CURRENT_TIME;
		return count-i;
	}

	/* If we fell out of the wait loop because of a signal, return
           ERESTARTSYS. --srhea */

	if (current->signal & ~current->blocked)
		return -ERESTARTSYS;

	/* Otherwise, we didn't read any bytes into the
           buffer. --srhea */

	return 0;
}

/*
 * TODO: comment me.
 */

static int aux_select(struct inode *inode, struct file *file, int sel_type, 
		      select_table * wait)
{
	if (sel_type != SEL_IN)
		return 0;
	if (aux_ready)
		return 1;
	select_wait(&queue->proc_list, wait);
	return 0;
}

/* Null file operations are ones this device does not support. --srhea */

struct file_operations psaux_fops = {
	NULL,		/* seek */
	read_aux,
	write_aux,
	NULL, 		/* readdir */
	aux_select,
	NULL, 		/* ioctl */
	NULL,		/* mmap */
	open_aux,
	release_aux,
	NULL,           /* fsync --srhea */
	fasync_aux,
                        /* check_media_change, revalidate NULL --srhea */
};


/*
 * Initialize driver. First check for a 82C710 chip; if found
 * forget about the Aux port and use the *_qp functions.
 */
static struct miscdevice psaux_mouse = {
	PSMOUSE_MINOR, "ps2aux", &psaux_fops
};

int psaux_init(void)
{
	int qp_found = 0;

	/* TODO: what is this all about? --srhea */

	if (aux_device_present == 0xaa) {
		printk(KERN_INFO "PS/2 auxiliary pointing device detected -- driver installed.\n");
	 	aux_present = 1;
		kbd_read_mask = AUX_OBUF_FULL;
	} else {
		return -EIO;
	}

	/* Register this driver with the kernel. --srhea */

	misc_register(&psaux_mouse);

	/* Allocate space for and set up the read queue. --srhea */

	queue = (struct aux_queue *) kmalloc(sizeof(*queue), GFP_KERNEL);
	memset(queue, 0, sizeof(*queue));
	queue->head = queue->tail = 0;
	queue->proc_list = NULL;

	/* Set the aux device up, then close it, since as of now no
           one has openned it. --srhea */

	if (!qp_found) {
#if defined INITIALIZE_DEVICE
		outb_p(AUX_ENABLE,AUX_COMMAND);		/* Enable Aux */
		aux_write_ack(AUX_SET_SAMPLE);
		aux_write_ack(100);			/* 100 samples/sec */
		aux_write_ack(AUX_SET_RES);
		aux_write_ack(3);			/* 8 counts per mm */
		aux_write_ack(AUX_SET_SCALE21);		/* 2:1 scaling */
		poll_aux_status_nosleep();
#endif /* INITIALIZE_DEVICE */
		outb_p(AUX_DISABLE,AUX_COMMAND);   /* Disable Aux device */
		poll_aux_status_nosleep();
		outb_p(AUX_CMD_WRITE,AUX_COMMAND);
		poll_aux_status_nosleep();             /* Disable interrupts */
		outb_p(AUX_INTS_OFF, AUX_OUTPUT_PORT); /*  on the controller */
	}
	return 0;
}

#ifdef MODULE
int init_module(void)
{
	return psaux_init(); /*??  Bjorn */
}

void cleanup_module(void)
{
	misc_deregister(&psaux_mouse);
	kfree(queue);
}
#endif

/* Is the device ready for a command? --srhea */

static int poll_aux_status(void)
{
	int retries=0;

	/* Until the input buffer full and output buffer full bits of
           the status register are zero, or until we have reached the
           maximum number of retries, sleep for a little while. See
           the hardware notes at the top of this file. --srhea */

	while ((inb(AUX_STATUS)&0x03) && retries < MAX_RETRIES) {

	  /* AUX_OBUF_FULL checks bits 5 and 0 of the status byte,
             which represent */

 		if ((inb_p(AUX_STATUS) & AUX_OBUF_FULL) == AUX_OBUF_FULL)
			inb_p(AUX_INPUT_PORT);
		current->state = TASK_INTERRUPTIBLE;
		current->timeout = jiffies + (5*HZ + 99) / 100;
		schedule();
		retries++;
	}

        /* Return true on success, false if retried too many
           times. --srhea */

	return !(retries==MAX_RETRIES);
}

static int poll_aux_status_nosleep(void)
{
	int retries = 0;

	while ((inb(AUX_STATUS)&0x03) && retries < 1000000) {
 		if ((inb_p(AUX_STATUS) & AUX_OBUF_FULL) == AUX_OBUF_FULL)
			inb_p(AUX_INPUT_PORT);
		retries++;
	}
	return !(retries == 1000000);
}
